"""
Module to get informations about WLAN devices.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Bernd Strebel, Klaus Bremer


import itertools
from ..core.exceptions import FritzServiceError
from ..core.fritzconnection import FritzConnection


SERVICE = 'WLANConfiguration'


class FritzWLAN:
    """
    Class to list all known wlan devices. All parameters are optional.
    If given, they have the following meaning: *fc* is an instance of
    FritzConnection, *address* the ip of the Fritz!Box, *port* the port
    to connect to, *user* the username, *password* the password. The
    *service* parameter specifies the configuration in use. Typically
    this is 1 for 2.4 GHz, 2 for 5 GHz and 3 for a guest network. This
    can vary depending on the router model and change with future
    standards.
    """

    def __init__(self, fc=None, address=None, port=None,
                       user=None, password=None, service=1):
        if fc is None:
            fc = FritzConnection(address, port, user, password)
        self.fc = fc
        self.service = service

    def _action(self, actionname, **kwargs):
        service = f'{SERVICE}{self.service}'
        return self.fc.call_action(service, actionname, **kwargs)

    @property
    def host_number(self):
        """
        Number of registered wlan devices for the active
        WLANConfiguration.
        """
        result = self._action('GetTotalAssociations')
        return result['NewTotalAssociations']

    @property
    def total_host_number(self):
        """
        Total NewAssociatedDeviceIndexumber of registered wlan devices
        for all WLANConfigurations.
        """
        total = 0
        _service = self.service
        for service in itertools.count(1):
            self.service = service
            try:
                total += self.host_number
            except FritzServiceError:
                break
        self.service = _service
        return total


    @property
    def ssid(self):
        """The WLAN SSID"""
        result = self._action('GetSSID')
        return result['NewSSID']

    @ssid.setter
    def ssid(self, value):
        self._action('SetSSID', NewSSID=value)

    @property
    def channel(self):
        """The WLAN channel in use"""
        return self.channel_infos()['NewChannel']

    @property
    def alternative_channels(self):
        """Alternative channels (as string)"""
        return self.channel_infos()['NewPossibleChannels']

    def channel_infos(self):
        """
        Return a dictionary with the keys *NewChannel* and
        *NewPossibleChannels* indicating the active channel and
        alternative ones.
        """
        return self._action('GetChannelInfo')

    def set_channel(self, number):
        """
        Set a new channel. *number* must be a valid channel number for
        the active WLAN. (Valid numbers are listed by *alternative_channels*.)
        """
        self._action('SetChannel', NewChannel=number)

    def get_generic_host_entry(self, index):
        """
        Return a dictionary with informations about the device
        internally stored at the position 'index'.
        """
        result = self._action(
            'GetGenericAssociatedDeviceInfo',
            NewAssociatedDeviceIndex=index
        )
        return result

    def get_specific_host_entry(self, mac_address):
        """
        Return a dictionary with informations about the device
        with the given 'mac_address'.
        """
        result = self._action(
            'GetSpecificAssociatedDeviceInfo',
            NewAssociatedDeviceMACAddress=mac_address
        )
        return result

    def get_hosts_info(self):
        """
        Returns a list of dictionaries with information about the known hosts.
        The dict-keys are: 'auth', 'mac', 'ip', 'signal', 'speed'
        """
        informations = []
        for index in itertools.count():
            try:
                host = self.get_generic_host_entry(index)
            except IndexError:
                break
            informations.append({
                'service': self.service,
                'index': index,
                'status': host['NewAssociatedDeviceAuthState'],
                'mac': host['NewAssociatedDeviceMACAddress'],
                'ip': host['NewAssociatedDeviceIPAddress'],
                'signal': host['NewX_AVM-DE_SignalStrength'],
                'speed': host['NewX_AVM-DE_Speed']
            })
        return informations



"""
Module to communicate with the AVM Fritz!Box.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import os
import string

from .devices import DeviceManager
from .exceptions import FritzServiceError
from .soaper import Soaper


# FritzConnection defaults:
FRITZ_IP_ADDRESS = '169.254.1.1'
FRITZ_TCP_PORT = 49000
FRITZ_IGD_DESC_FILE = 'igddesc.xml'
FRITZ_TR64_DESC_FILE = 'tr64desc.xml'
FRITZ_USERNAME = 'dslf-config'


class FritzConnection:
    """
    Main class to set up a connection to the Fritz!Box router. All
    parameters are optional. `address` should be the ip of a router, in
    case that are multiple Fritz!Box routers in a network, the ip must
    be given. Otherwise it is undefined which router will respond. If
    `user` and `password` are not provided, the environment gets checked for
    FRITZ_USERNAME and FRITZ_PASSWORD settings and taken from there, if
    found.

    The optional parameter `timeout` is a floating number in seconds
    limiting the time waiting for a router response. This is a global
    setting for the internal communication with the router. In case of a
    timeout a `requests.ConnectTimeout` exception gets raised.
    (`New in version 1.1`)
    """

    def __init__(self, address=None, port=None, user=None, password=None,
                       timeout=None):
        """
        Initialisation of FritzConnection: reads all data from the box
        and also the api-description (the servicenames and according
        actionnames as well as the parameter-types) that can vary among
        models and stores these informations as instance-attributes.
        This can be an expensive operation. Because of this an instance
        of FritzConnection should be created once and reused in an
        application. All parameters are optional. But if there is more
        than one FritzBox in the network, an address (ip as string) must
        be given, otherwise it is not defined which box may respond. If
        no user is given the Environment gets checked for a
        FRITZ_USERNAME setting. If there is no entry in the environment
        the avm-default-username will be used. If no password is given
        the Environment gets checked for a FRITZ_PASSWORD setting. So
        password can be used without using configuration-files or even
        hardcoding. The optional parameter `timeout` is a floating
        number in seconds limiting the time waiting for a router
        response. This is a global setting for the internal
        communication with the router. In case of a timeout a
        `requests.ConnectTimeout` exception gets raised.
        """
        if address is None:
            address = FRITZ_IP_ADDRESS
        if port is None:
            port = FRITZ_TCP_PORT
        if user is None:
            user = os.getenv('FRITZ_USERNAME', FRITZ_USERNAME)
        if password is None:
            password = os.getenv('FRITZ_PASSWORD', '')

        self.soaper = Soaper(address, port, user, password, timeout=timeout)
        self.device_manager = DeviceManager(timeout=timeout)

        descriptions = [FRITZ_IGD_DESC_FILE]
        if password:
            descriptions.append(FRITZ_TR64_DESC_FILE)
        for description in descriptions:
            source = f'http://{address}:{port}/{description}'
            try:
                self.device_manager.add_description(source)
            except OSError:
                # resource not available: ignore this
                pass

        self.device_manager.scan()
        self.device_manager.load_service_descriptions(address, port)

    def __repr__(self):
        """Return a readable representation"""
        return  f"{self.modelname} at ip {self.soaper.address}\n"\
                f"FRITZ!OS: {self.system_version}"

    @property
    def services(self):
        """
        Dictionary of service instances. Keys are the service names.
        """
        return self.device_manager.services

    @property
    def modelname(self):
        """
        Returns the modelname of the router.
        """
        return self.device_manager.modelname

    @property
    def system_version(self):
        """
        Returns system version if known, otherwise None.
        """
        return self.device_manager.system_version

    @staticmethod
    def normalize_name(name):
        """
        Returns the normalized service name. E.g. WLANConfiguration and
        WLANConfiguration:1 will get converted to WLANConfiguration1.
        """
        if ':' in name:
            name, number = name.split(':', 1)
            name = name + number
        elif name[-1] not in string.digits:
            name = name + '1'
        return name


    # -------------------------------------------
    # public api:
    # -------------------------------------------

    def call_action(self, service_name, action_name, *,
                    arguments=None, **kwargs):
        """
        Executes the given action of the given service. Both parameters
        are required. Arguments are optional and can be provided as a
        dictionary given to 'arguments' or as separate keyword
        parameters. If 'arguments' is given additional
        keyword-parameters as further arguments are ignored.
        If the service_name does not end with a number (like 1), a 1
        gets added by default. If the service_name ends with a colon and a
        number, the colon gets removed. So i.e. WLANConfiguration
        expands to WLANConfiguration1 and WLANConfiguration:2 converts
        to WLANConfiguration2.
        Invalid service names will raise a ServiceError and invalid
        action names will raise an ActionError.
        """
        arguments = arguments if arguments else dict()
        if not arguments:
            arguments.update(kwargs)
        service_name = self.normalize_name(service_name)
        try:
            service = self.device_manager.services[service_name]
        except KeyError:
            raise FritzServiceError(f'unknown service: "{service_name}"')
        return self.soaper.execute(service, action_name, arguments)

    def reconnect(self):
        """
        Terminate the connection and reconnects with a new ip.
        """
        self.call_action('WANIPConn1', 'ForceTermination')



import pytest

from ..lib.fritztools import format_num, format_rate


@pytest.mark.parametrize(
    "num, formated_num", [
        (300, '300.0 bytes'),
        (2048, '2.0 KB'),
        (3000, '2.9 KB'),
        (32500, '31.7 KB'),
        (45e6, '42.9 MB'),
        (45e9, '41.9 GB'),
        (45e12, '40.9 TB'),
        (45e15, '40.0 PB'),
    ]
)

def test_format_num(num, formated_num):
    assert formated_num == format_num(num)




"""
Implements the DeviceManager for physical and virtual devices. Every
physical device (a router) has a set of virtual subdevices.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from .processor import Description
from .utils import get_xml_root


class DeviceManager:
    """
    Knows all data about the device and the subdevices, including the
    available services. Takes an optional `timeout` parameter to limit the time waiting for a router response.
    """

    def __init__(self, timeout=None):
        self.descriptions = []
        self.services = {}
        self.timeout = timeout

    @property
    def modelname(self):
        """
        Take the root-device of the first description and return the
        according modelname. This is the name of the Fritz!Box itself.
        Will raise an IndexError if the method is called before
        descriptions are added.
        """
        return self.descriptions[0].device_model_name

    @property
    def system_version(self):
        """
        Returns a tuple with version, display and buildnumber from the
        first description providing this informations. Returns None if
        no system informations are available.
        """
        version = None
        for description in self.descriptions:
            version = description.system_version
            if version:
                return version
        return None

    def add_description(self, source):
        """
        Adds description data about the devices and the according
        services. 'source' is a string with the xml-data, like the
        content of an igddesc- or tr64desc-file.
        """
        root = get_xml_root(source, timeout=self.timeout)
        self.descriptions.append(Description(root))

    def scan(self):
        """
        Scans all available services defined by the description files.
        Must get called after all xml-descriptions are added.
        """
        for description in self.descriptions:
            self.services.update(description.services)

    def load_service_descriptions(self, address, port):
        """
        Triggers the load of the scpd files of the services, so they
        known their actions.
        """
        for service in self.services.values():
            service.load_scpd(address, port, timeout=self.timeout)




"""
Modul to list the known hosts. Older versions of FritzOS lists only up
to 16 entries. For newer versions this limitation is gone.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import itertools
from ..core.fritzconnection import FritzConnection


SERVICE = 'Hosts'


class FritzHosts:
    """
    Class to list all known hosts. All parameters are optional. If
    given, they have the following meaning: *fc* is an instance of
    FritzConnection, *address* the ip of the Fritz!Box, *port* the port
    to connect to, *user* the username, *password* the password.
    """

    def __init__(self, fc=None, address=None, port=None, user=None, password=None):
        super().__init__()
        if fc is None:
            fc = FritzConnection(address, port, user, password)
        self.fc = fc

    def _action(self, actionname, **kwargs):
        return self.fc.call_action(SERVICE, actionname, **kwargs)

    @property
    def modelname(self):
        """The router modelname."""
        return self.fc.modelname

    @property
    def host_numbers(self):
        """The number of known hosts."""
        result = self._action('GetHostNumberOfEntries')
        return result['NewHostNumberOfEntries']

    def get_generic_host_entry(self, index):
        """
        Returns a dictionary with informations about a device internally
        registered by the position *index*. Index-positions are
        zero-based.
        """
        result = self._action('GetGenericHostEntry', NewIndex=index)
        return result

    def get_specific_host_entry(self, mac_address):
        """
        Returns a dictionary with informations about a device addressed
        by the MAC-address.
        """
        result = self._action('GetSpecificHostEntry', NewMACAddress=mac_address)
        return result

    def get_hosts_info(self):
        """
        Returns a list of dicts with information about the known hosts.
        The dict-keys are: 'ip', 'name', 'mac', 'status'
        """
        result = []
        for index in itertools.count():
            try:
                host = self.get_generic_host_entry(index)
            except IndexError:
                # no more host entries:
                break
            result.append({
                'ip': host['NewIPAddress'],
                'name': host['NewHostName'],
                'mac': host['NewMACAddress'],
                'status': host['NewActive']})
        return result





import argparse

from .. import package_version
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)
from ..lib.fritzcall import FritzCall


def get_fritzcall(arguments):
    return FritzCall(address=arguments.address,
                     port=arguments.port,
                     user=arguments.username,
                     password=arguments.password)


def report_calls(fc, arguments):
    print(f'\nFritzConnection v{package_version}')
    print(fc.fc)
    days = arguments.days
    num = arguments.num if not days else None
    if arguments.type == 'in':
        calls = fc.get_received_calls(num=num, days=days)
    elif arguments.type == 'out':
        calls = fc.get_out_calls(num=num, days=days)
    elif arguments.type == 'missed':
        calls = fc.get_missed_calls(num=num, days=days)
    else:
        calls = fc.get_calls(num=num, days=days)
    call_type = arguments.type if arguments.type else 'all'
    print('\nList of calls:', call_type, '\n')
    type_ = 'type'
    number = 'number'
    time = 'date/time'
    duration = 'duration'
    print(f'{type_:>6}   {number:24}{time:>18}{duration:>12}\n')
    for call in calls:
        print(call)


def dial_number(fc, number):
    print(f'dialing number: {number}')
    fc.dial(number)
    print('dialing done, please wait for signal.')


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='FritzBox Callhistory')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help='ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help='port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    parser.add_argument('-n', '--num',
                        nargs='?', default=None, const=None,
                        help='max number of calls in the call-list')
    parser.add_argument('-d', '--days',
                        nargs='?', default=None, const=None,
                        help='number of days to look back for calls.')
    parser.add_argument('-t', '--type',
                        nargs='?', default=None, const=None,
                        help='type of calls: [in|out|missed]')
    parser.add_argument('-c', '--call',
                        nargs='?', default=None, const=None,
                        help='phone number to call')
    args = parser.parse_args()
    return args


def main():
    arguments = get_cli_arguments()
    if not arguments.password:
        print('Exit: password required.')
        return
    fc = get_fritzcall(arguments)
    if arguments.call:
        dial_number(fc, arguments.call)
    else:
        report_calls(fc, arguments)


if __name__ == '__main__':
    main()




import os
import pytest

from ..core.processor import (
    Description,
    Device,
    Service,
    Scpd,
)
from ..core.utils import get_xml_root


@pytest.fixture()
def igddesc_source(datadir):
    os.chdir(datadir)
    return get_xml_root('igddesc.xml')


@pytest.fixture()
def tr64desc_source(datadir):
    os.chdir(datadir)
    return get_xml_root('tr64desc.xml')


@pytest.fixture()
def igdconn_scpd_source(datadir):
    os.chdir(datadir)
    return get_xml_root('igdconnSCPD.xml')


@pytest.fixture()
def scpd_instance(igdconn_scpd_source):
    return Scpd(igdconn_scpd_source)


@pytest.mark.parametrize(
    "attribute, value", [
        ('spec_version', '1.0'),
        ('device_model_name', 'FRITZ!Box 7590'),
    ])
def test_igd_description_specversion(attribute, value, igddesc_source):
    d = Description(igddesc_source)
    assert d.spec_version == '1.0'


@pytest.mark.parametrize(
    "attribute, value", [
        ('deviceType', 'urn:schemas-upnp-org:device:InternetGatewayDevice:1'),
        ('friendlyName', 'FRITZ!Box 7590'),
        ('manufacturer', 'AVM Berlin'),
        ('manufacturerURL', 'http://www.avm.de'),
        ('modelDescription', 'FRITZ!Box 7590'),
        ('modelName', 'FRITZ!Box 7590'),
        ('modelNumber', 'avm'),
        ('modelURL', 'http://www.avm.de'),
        ('UDN', 'uuid:75802409-bccb-just-a-test-udn'),
    ])
def test_igd_device(attribute, value, igddesc_source):
    d = Description(igddesc_source)
    device = d.device
    assert getattr(device, attribute) == value


def test_igd_device_services(igddesc_source):
    # the device should have one service instance in services.
    d = Description(igddesc_source)
    services = d.device._services
    assert len(services) == 1
    assert isinstance(services[0], Service)


@pytest.mark.parametrize(
    "attribute, value", [
        ('serviceType', 'urn:schemas-any-com:service:Any:1'),
        ('serviceId', 'urn:any-com:serviceId:any1'),
        ('controlURL', '/igdupnp/control/any'),
        ('eventSubURL', '/igdupnp/control/any'),
        ('SCPDURL', '/any.xml'),
    ])
def test_igd_device_service(attribute, value, igddesc_source):
    d = Description(igddesc_source)
    service = d.device._services[0]
    assert getattr(service, attribute) == value


def test_igd_device_service_name(igddesc_source):
    """
    Every service has a name. First (and only) service of the main
    device has the name 'any1'.
    """
    d = Description(igddesc_source)
    service = d.device._services[0]
    assert service.name == 'any1'


def test_igd_device_devices(igddesc_source):
    # the device should have one device instance in devices.
    d = Description(igddesc_source)
    devices = d.device.devices
    assert len(devices) == 1
    assert isinstance(devices[0], Device)


def test_idg_nested_services_part1(igddesc_source):
    """
    The main device should have a sub-device with another subdevice. The
    first sub-device should have a sublist with one service.
    """
    d = Description(igddesc_source)
    sub_device = d.device.devices[0]
    assert len(sub_device._services) == 1


def test_idg_nested_services_part2(igddesc_source):
    """
    The main device should have a sub-device with another subdevice. The
    sub-sub-device should have a sublist with three service.
    """
    d = Description(igddesc_source)
    sub_device = d.device.devices[0].devices[0]
    assert len(sub_device._services) == 3


@pytest.mark.parametrize(
    "name", [
        'WANCommonIFC1', 'WANDSLLinkC1', 'WANIPConn1'
    ])
def test_igd_description_services(name, igddesc_source):
    """
    Find services by name from nested devices (igddesc.xml)
    """
    d = Description(igddesc_source)
    service = d.services[name]
    assert service.name == name


@pytest.mark.parametrize(
    "name", [
        'DeviceInfo1', 'X_AVM-DE_Homeauto1', 'WLANConfiguration1',
        'WANCommonInterfaceConfig1', 'WANIPConnection1',
    ])
def test_tr64_description_services(name, tr64desc_source):
    """
    Find services by name from nested devices (tr64desc.xml)
    """
    d = Description(tr64desc_source)
    service = d.services[name]
    assert service.name == name


def test_system_version_igd(igddesc_source):
    d = Description(igddesc_source)
    assert d.system_version == None


def test_system_version_tr64(tr64desc_source):
    d = Description(tr64desc_source)
    assert d.system_version == '7.10'


def test_system_buildnumber_tr64(tr64desc_source):
    d = Description(tr64desc_source)
    assert d.system_buildnumber == '67453'


def test_scpd_specversion(scpd_instance):
    assert scpd_instance.spec_version == '1.0'


def test_scpd_actionlist(scpd_instance):
    # 18 actions are defined by 'igdconnSCPD.xml':
    assert len(scpd_instance._actions) == 18


def test_scpd_actiondict(scpd_instance):
    # 18 actions are defined by 'igdconnSCPD.xml':
    assert len(scpd_instance.actions) == 18


@pytest.mark.parametrize(
    "name", [
        'SetConnectionType', 'ForceTermination', 'X_AVM_DE_GetIPv6DNSServer'
    ])
def test_scpd_action_names(name, scpd_instance):
    # find some action names
    actions = scpd_instance.actions
    print(actions)
    action = actions[name]
    assert action.name == name


@pytest.mark.parametrize(
    "action_name, argument_name, direction, relatedStateVariable", [
        ('GetStatusInfo', 'NewConnectionStatus', 'out', 'ConnectionStatus'),
        ('GetStatusInfo', 'NewLastConnectionError', 'out', 'LastConnectionError'),
        ('GetStatusInfo', 'NewUptime', 'out', 'Uptime')
    ])
def test_scpd_action_argument_names(
        action_name, argument_name, direction,
        relatedStateVariable, scpd_instance
    ):
    """
    Test action arguments
    """
    action = scpd_instance.actions[action_name]
    argument = action.arguments[argument_name]
    assert argument.name == argument_name
    assert argument.direction == direction
    assert argument.relatedStateVariable == relatedStateVariable


def test_scpd_service_state_variables(scpd_instance):
    # 30 stateVariables are defined by 'igdconnSCPD.xml':
    assert len(scpd_instance._state_variables) == 30


@pytest.mark.parametrize(
    "name, data_type, default_value, allowed_values_num", [
        ('LastConnectionError', 'string', 'ERROR_NONE', 23),
        ('AutoDisconnectTime', 'ui4', '0', 0),
    ])
def test_scpd_service_state_variables_by_name(
        name, data_type, default_value, allowed_values_num, scpd_instance
    ):
    sv = scpd_instance.state_variables[name]
    assert sv.name == name
    assert sv.dataType == data_type
    assert sv.defaultValue == default_value
    assert len(sv.allowed_values) == allowed_values_num


def test_scpd_statevariable_allowed_value(scpd_instance):
    sv = scpd_instance.state_variables['LastConnectionError']
    assert 'ERROR_NO_CARRIER' in sv.allowed_values


def test_scpd_statevariable_allowed_range(scpd_instance):
    sv = scpd_instance.state_variables['Uptime']
    allowed_range = sv.allowedValueRange
    assert allowed_range.minimum == '0'
    assert allowed_range.maximum == '4294967295'
    assert allowed_range.step == '1'



"""
Module for read-only access to the contents of the Fritz!Box phonebooks.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Authors: Klaus Bremer, David M. Straub


from ..core.fritzconnection import FritzConnection
from ..core.processor import (
    processor,
    process_node,
    InstanceAttributeFactory,
    Storage,
    ValueSequencer,
)
from ..core.utils import get_xml_root


__all__ = ['FritzPhonebook1']


SERVICE = 'X_AVM-DE_OnTel'


class FritzPhonebook(object):
    """
    Interface to access the Fritz!Box phonebooks. All parameters are
    optional. If given, they have the following meaning: `fc` is an
    instance of FritzConnection, `address` the ip of the Fritz!Box,
    `port` the port to connect to, `user` the username, `password` the
    password.
    """
    def __init__(self, fc=None, address=None, port=None,
                       user=None, password=None):
        if fc is None:
            fc = FritzConnection(address, port, user, password)
        self.fc = fc
        self.phonebook = None

    def _action(self, actionname, **kwargs):
        return self.fc.call_action(SERVICE, actionname, **kwargs)

    @property
    def modelname(self):
        """
        The router modelname.
        """
        return self.fc.modelname

    @property
    def phonebook_ids(self):
        """
        List of integers identifying the phonebooks. This property is
        defined as `phonebook_ids` and as `list_phonebooks` for backward
        compatibility. The property `list_phonebooks` is deprecated and
        may get removed in the future.
        """
        result = self._action('GetPhonebookList')
        try:
            res = result['NewPhonebookList'].split(',')
            res = [int(x) for x in res]
        except KeyError:
            return []
        return res

    # legathy api name for backward compatibility
    list_phonebooks = phonebook_ids

    def phonebook_info(self, id):
        """
        Get the `name`, `url` and an optional `extra id` of the
        phonebook with integer `id`. Returns a dictionary with the keys
        `name`, `url` and `xid`.
        """
        result = self._action('GetPhonebook', NewPhonebookId=id)
        return {
            'name': result.get('NewPhonebookName'),
            'url': result.get('NewPhonebookURL'),
            'xid': result.get('NewPhonebookExtraID')
        }

    def get_all_names(self, id):
        """
        Get a dictionary with all names and their phone numbers for the
        phonebook with `id`.
        """
        url = self.phonebook_info(id)['url']
        self._read_phonebook(url)
        return {
            contact.name: contact.numbers
            for contact in self.phonebook.contacts
        }

    def get_all_numbers(self, id):
        """
        Get a dictionary with all phone numbers and the according names
        for the phonebook with `id`.
        """
        reverse_contacts = dict()
        for name, numbers in self.get_all_names(id).items():
            for number in numbers:
                reverse_contacts[number] = name
        return reverse_contacts

    def lookup_numbers(self, id, name):
        """
        Look up the phone numbers of contact `name` in the phonebook
        with `id`. Returns a list of numbers. Will raise a KeyError if
        the name is unknown.
        """
        return self.get_all_names(id)[name]

    def lookup_names(self, id, number):
        """
        Look up the names of the contacts with phone number `number` in
        the phonebook with `id`. Will raise a KeyError if the number is
        unknown.
        """
        return self.get_all_numbers(id)[number]

    def _read_phonebook(self, url):
        """
        Read the content of the phonebook with the given `url`. This
        method sets the phone book instance attribute and has no return
        value.
        """
        root = get_xml_root(url)
        self.phonebook = Phonebook()
        process_node(self, root)


@processor
class Services:
    """
    Services container. So far just for an associated email-address.
    """
    def __init__(self):
        self.email = None


@processor
class Person:
    """
    Data storage for a contact name and an image.
    """
    def __init__(self):
        self.realName = None
        self.imageURL = None


@processor
class Telephony:
    """
    Data storage for the phone numbers of the contact and services.
    """
    number = ValueSequencer('numbers')

    def __init__(self):
        self.numbers = list()
        self.services = Services()


@processor
class Contact:
    """
    Represents a contact with `catecory`- and `uniqueid`-attributes as
    well as `person`- and telephony-subnodes.
    """

    def __init__(self):
        self.category = None
        self.uniqueid = None
        self.person = Person()
        self.telephony = Telephony()

    @property
    def name(self):
        return self.person.realName

    @property
    def numbers(self):
        return self.telephony.numbers


@processor
class Phonebook(Storage):
    """
    Represents a phonebook with a timestamp indicating the last
    modification and a list of Contact instances.
    """
    contact = InstanceAttributeFactory(Contact)

    def __init__(self):
        self.timestamp = None
        self.contacts = list()
        super().__init__(self.contacts)



"""
Modul to read status-informations from an AVM FritzBox.
"""

import time

from ..core.fritzconnection import FritzConnection
from ..core.exceptions import FritzServiceError
from .fritztools import format_num, format_rate


class FritzStatus:
    """
    Class for requesting status-informations:
    up, down, ip, activity (bytes per second send/received).
    All parameters are optional. If given, they have the following meaning:
    *fc* is an instance of FritzConnection,
    *address* the ip of the Fritz!Box,
    the *port* of the Fritz!Box and the according *user* and *password*.
    """

    def __init__(self, fc=None, address=None, port=None,
                       user=None, password=None):
        """
        :fc: instance of FritzConnection.
        :address: ip of the Fritz!Box
        """
        self.fc = fc if fc else FritzConnection(address, port, user, password)
        # depending on the model (i.e. a repeater) some services
        # may not be available. Don't let FritzStatus crash at init.
        try:
            self.last_bytes_sent = self.bytes_sent
            self.last_bytes_received = self.bytes_received
        except FritzServiceError:
            pass
        self.last_traffic_call = time.time()

    @property
    def modelname(self):
        """The router modelname."""
        return self.fc.modelname

    @property
    def is_linked(self):
        """
        A boolean whether the FritzBox is physically linked to
        the provider.
        """
        status = self.fc.call_action('WANCommonIFC',
                                     'GetCommonLinkProperties')
        return status['NewPhysicalLinkStatus'] == 'Up'

    @property
    def is_connected(self):
        """
        A boolean whether the FritzBox has established an
        internet-connection.
        """
        status = self.fc.call_action('WANIPConn', 'GetStatusInfo')
        return status['NewConnectionStatus'] == 'Connected'

    @property
    def external_ip(self):
        """The external v4 ip-address."""
        return self.fc.call_action('WANIPConn',
            'GetExternalIPAddress')['NewExternalIPAddress']

    @property
    def external_ipv6(self):
        """The external v6 ip-address."""
        return self.fc.call_action('WANIPConn',
            'X_AVM_DE_GetExternalIPv6Address')['NewExternalIPv6Address']

    @property
    def uptime(self):
        """Uptime in seconds."""
        status = self.fc.call_action('WANIPConn', 'GetStatusInfo')
        return status['NewUptime']

    @property
    def str_uptime(self):
        """Uptime in seconds and in human readable format."""
        mins, secs = divmod(self.uptime, 60)
        hours, mins = divmod(mins, 60)
        return '%02d:%02d:%02d' % (hours, mins, secs)

    @property
    def bytes_sent(self):
        """Total number of send bytes."""
        status = self.fc.call_action('WANCommonIFC',
                                     'GetTotalBytesSent')
        return status['NewTotalBytesSent']

    @property
    def bytes_received(self):
        """Total number of received bytes."""
        status = self.fc.call_action('WANCommonIFC',
                                     'GetTotalBytesReceived')
        return status['NewTotalBytesReceived']

    @property
    def transmission_rate(self):
        """
        The upstream and downstream values as a tuple in bytes per
        second. Use this for periodical calling.
        """
        sent = self.bytes_sent
        received = self.bytes_received
        traffic_call = time.time()
        time_delta = traffic_call - self.last_traffic_call
        upstream = int(1.0 * (sent - self.last_bytes_sent)/time_delta)
        downstream = int(1.0 * (received - self.last_bytes_received)/time_delta)
        self.last_bytes_sent = sent
        self.last_bytes_received = received
        self.last_traffic_call = traffic_call
        return upstream, downstream

    @property
    def str_transmission_rate(self):
        """
        Tuple of human readable transmission rate in bytes. First item
        is upstream, second item downstream.
        """
        upstream, downstream = self.transmission_rate
        return (
            format_num(upstream),
            format_num(downstream)
        )

    @property
    def max_linked_bit_rate(self):
        """
        Tuple with the maximun upstream- and downstream-rate
        of the physical link. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate('WANCommonInterfaceConfig')

    @property
    def max_bit_rate(self):
        """
        Tuple with the maximun upstream- and downstream-rate
        of the given connection. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate('WANCommonIFC')

    def _get_max_bit_rate(self, servicename):
        """
        internal method to get the upstream and downstream-rates for
        different services of the WANCommonInterfaceConfig1 ServiceType.
        """
        status = self.fc.call_action(servicename, 'GetCommonLinkProperties')
        downstream = status['NewLayer1DownstreamMaxBitRate']
        upstream = status['NewLayer1UpstreamMaxBitRate']
        return upstream, downstream

    @property
    def max_byte_rate(self):
        """
        Same as max_bit_rate but rate is given in bytes/sec.
        """
        upstream, downstream = self.max_bit_rate
        return upstream / 8.0, downstream / 8.0

    @property
    def str_max_linked_bit_rate(self):
        """
        Human readable maximum of the physical upstream- and
        downstream-rate in bits/sec. Value is a tuple, first item is
        upstream, second item is downstream.
        """
        upstream, downstream = self.max_linked_bit_rate
        return (
            format_rate(upstream, unit='bits'),
            format_rate(downstream, unit ='bits')
        )

    @property
    def str_max_bit_rate(self):
        """
        Human readable maximum of the upstream- and downstream-rate in
        bits/sec, as given by the provider. Value is a tuple, first item
        is upstream, second item is downstream.
        """
        upstream, downstream = self.max_bit_rate
        return (
            format_rate(upstream, unit='bits'),
            format_rate(downstream, unit ='bits')
        )

    def reconnect(self):
        """Makes a reconnection with a new external ip."""
        self.fc.reconnect()



"""
fritzstatus.py

Module to inspect the FritzBox API for available services and actions.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

import argparse

from .. import package_version
from ..core.exceptions import (
    FritzServiceError,
    FritzActionError,
)
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)
from ..lib.fritzstatus import FritzStatus


def print_status(address=None, port=None, user=None, password=None):
    print(f'\nFritzConnection v{package_version}')
    fs = FritzStatus(address=address, port=port, user=user, password=password)
    print(f'FritzStatus for {fs.fc}:\n')
    status_informations = [
        ('is linked', 'is_linked'),
        ('is connected', 'is_connected'),
        ('external ip (v4)', 'external_ip'),
        ('external ip (v6)', 'external_ipv6'),
        ('uptime', 'str_uptime'),
        ('bytes send', 'bytes_sent'),
        ('bytes received', 'bytes_received'),
        ('max. bit rate', 'str_max_bit_rate'),
    ]
    for status, attribute in status_informations:
        try:
            information = getattr(fs, attribute)
        except (FritzServiceError, FritzActionError):
            information = f'unsupported attribute "{attribute}"'
        print(f'    {status:20}: {information}')


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='FritzBox Status')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help=f'ip-address of the FritzBox to connect to. '
                             f'Default: {FRITZ_IP_ADDRESS}')
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help=f'port of the FritzBox to connect to. '
                             f'Default: {FRITZ_TCP_PORT}')
    args = parser.parse_args()
    return args


def main():
    args = get_cli_arguments()
    if not args.password:
        print('Exit: password required.')
    else:
        print_status(args.address, args.port, args.username, args.password)


if __name__ == '__main__':
    main()




import argparse
import itertools

from ..lib.fritzwlan import (
    FritzWLAN,
    SERVICE,
)
from ..core.exceptions import FritzServiceError
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)


def get_header():
    index = 'index'
    status = 'active'
    mac = 'mac'
    ip = 'ip'
    signal = 'signal'
    speed = 'speed'
    return f'{index:>5}{status:>8}{mac:>20}{ip:>18}{signal:>8}{speed:>8}'


def report_wlanconfiguration(fw, extension):
    fw.service = extension
    host_informations = fw.get_hosts_info()
    if host_informations:
        print(f'Hosts registered at {SERVICE}{extension}:')
        print(f'WLAN name: {fw.ssid}')
        print(f'channel  : {fw.channel}')
        print(get_header())
        for info in host_informations:
            index = info['index']
            status = info['status']
            mac = info['mac']
            ip = info['ip']
            signal = info['signal']
            speed = info['speed']
            print(f'{index:>5}{status:>8}{mac:>20}{ip:>18}{signal:>8}{speed:>8}')
        print()


def report_devices(arguments):
    fw = FritzWLAN(address=arguments.address,
                   port=arguments.port,
                   user=arguments.username,
                   password=arguments.password,
                   service=arguments.service)
    print(fw.fc)
    if arguments.service:
        try:
            report_wlanconfiguration(fw, arguments.service)
        except FritzServiceError as err:
            print(f'Error: {err}')
    else:
        for n in itertools.count(1):
            try:
                report_wlanconfiguration(fw, n)
            except FritzServiceError:
                break


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='FritzBox HomeAuto')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help='ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help='port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    parser.add_argument('-s', '--service',
                        nargs='?', default=0, const=None,
                        help='WLANConfiguration service number')
    args = parser.parse_args()
    return args


def main():
    arguments = get_cli_arguments()
    if not arguments.password:
        print('Exit: password required.')
    else:
        report_devices(arguments)


if __name__ == '__main__':
    main()



"""
fritzphonebook.py

Module to inspect the Fritz!Box phonebooks.
CLI interface.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Authors: Klaus Bremer, David M. Straub


import argparse

from .. import package_version
from ..lib.fritzphonebook import FritzPhonebook
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)


def print_header(fpb):
    print(f'\nFritzConnection v{package_version}')
    print(f'FritzPhonebook for {fpb.fc}\n')


def print_phonebooks(fpb):
    for id in fpb.phonebook_ids:
        info = fpb.phonebook_info(id)
        print(f"Content of phonebook: {info['name']} ")
        for name, numbers in fpb.get_all_names(id).items():
            print(f"{name:<30}{', '.join(numbers)}")
        print()


def print_search_name(fpb, arguments):
    found = False
    for id in fpb.phonebook_ids:
        contacts = fpb.get_all_names(id)
        numbers = contacts.get(arguments.name)
        if numbers:
            print(f"{arguments.name:<30}{', '.join(numbers)}")
            found = True
    if not found:
        print(f"name {arguments.name} not found.")


def print_search_number(fpb, arguments):
    found = False
    for id in fpb.phonebook_ids:
        numbers = fpb.get_all_numbers(id)
        if arguments.number in numbers:
            print(f"{numbers[arguments.number]:<30}{arguments.number}")
            found = True
    if not found:
        print(f'number {arguments.number} not found.')


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='Fritz!Box Phonebook')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help='ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help='port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    parser.add_argument('-a', '--all',
                        action='store_true',
                        help='List all phone books ')
    parser.add_argument('--name',
                        nargs='?', default=0,
                        help='Name for number search')
    parser.add_argument('--number',
                        nargs='?', default=0,
                        help='Number for name search')
    args = parser.parse_args()
    return args


def main():
    arguments = get_cli_arguments()
    if not arguments.password:
        print('Exit: password required.')
        return
    fpb = FritzPhonebook(address=arguments.address,
                         port=arguments.port,
                         user=arguments.username,
                         password=arguments.password)
    print_header(fpb)
    if arguments.all:
        print_phonebooks(fpb)
    elif arguments.name:
        print_search_name(fpb, arguments)
    elif arguments.number:
        print_search_number(fpb, arguments)
    print()  # blank line for better readability


if __name__ == '__main__':
    main()



"""
Module to parse and store the device description and the service data
provided by xml and the scpd protocol based on xml. Names partly violate
PEP8 representing node-names from xml description files.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from .utils import (
    get_xml_root,
    localname,
)


# ---------------------------------------------------------
# processor, decorator and descriptors here:
# ---------------------------------------------------------

def process_node(obj, root):
    """
    Take an object and a root of nodes. The node.text of nodes with the
    same name as an instance-attribute of 'obj' are set as values for
    the corresponding instance-attribute. If the attribute is a
    callable, processing the node gets delegated to the callable (which
    in turn calls process_node).
    """
    for node in root:
        node_name = localname(node)
        try:
            attr = getattr(obj, node_name)
        except AttributeError:
            # ignore unknown nodes
            continue
        if callable(attr):
            # attribute is a callable: delegate further
            attr(node)
        else:
            # node is an attribute: set value
            if isinstance(node.text, str):
                value = node.text.strip()
            else:
                value = node.text
            setattr(obj, node_name, value)


def processor(cls):
    """
    Class decorator to add the functionality of calling 'process_node'
    on invoking an instance as a callable.
    """
    cls.__call__ = lambda obj, root: process_node(obj, root)
    return cls


class ValueSequencer:
    """
    Data descriptor storing a value (assigned as attribute value) in a
    given sequence.
    """
    def __init__(self, sequence_name):
        self.sequence_name = sequence_name

    def __get__(self, obj, objtype):
        # kind of set only attribute
        return NotImplemented

    def __set__(self, obj, value):
        sequence = getattr(obj, self.sequence_name)
        sequence.append(value)


class InstanceAttributeFactory:
    """
    Non data descriptor returning instances of 'cls' and registering
    these instances in the '_storage' attribute of the calling instance.
    """
    def __init__(self, cls):
        self.cls = cls

    def __get__(self, obj, objtype):
        instance = self.cls()
        obj._storage.append(instance)
        return instance


class Storage:
    """
    Baseclass for classes working with InstanceAttributeFactory.
    """
    def __init__(self, storage):
        self._storage = storage


# ---------------------------------------------------------
# Node- and root-classes. Root classes are entry-points.
# Root classes are Scpd and Description.
# Description is the only public class.
# ---------------------------------------------------------

@processor
class SpecVersion:
    """
    Specification version from the schema device or service
    informations.
    """
    def __init__(self):
        # attributes are case sensitive node names:
        self.major = None
        self.minor = None

    @property
    def version(self):
        return f'{self.major}.{self.minor}'


@processor
class SystemVersion:
    """
    Information about the Fritz!OS version of the Fritz!Box.
    Information is just provided by the 'tr64desc.xml' file.
    """
    def __init__(self):
        # attributes are case sensitive node names
        self.HW = None
        self.Major = None
        self.Minor = None
        self.Patch = None
        self.Buildnumber = None
        self.Display = None

    @property
    def version(self):
        """
        Returns system version as string like '7.10' or None if system
        version is unknown.
        """
        if self.Minor and self.Patch:
            return f'{self.Minor}.{self.Patch}'
        return None


@processor
class Argument:
    """
    An argument with name, direction and relatedStateVariable
    attributes.
    """
    def __init__(self):
        # attributes are case sensitive node names
        self.name = None
        self.direction = None
        self.relatedStateVariable = None


@processor
class ArgumentList(Storage):
    """
    Collects the arguments for an action.
    """
    # case sensitive node
    argument = InstanceAttributeFactory(Argument)


@processor
class Action:
    """
    Every Action has a name and a list of arguments.
    """
    def __init__(self):
        self._arguments = list()
        self._arguments_storage = None
        # attributes are case sensitive node names:
        self.name = None
        self.argumentList = ArgumentList(self._arguments)

    @property
    def arguments(self):
        """
        Returns the action-arguments as a dict. argument-names are the
        keys and the argument objects are the values. The dictionary
        gets cached.
        """
        if not self._arguments_storage:
            self._arguments_storage = {arg.name: arg for arg in self._arguments}
        return self._arguments_storage


@processor
class ActionList(Storage):
    """
    Collection of actions of a service.
    The Action instances are stored in the Scpd.actions attribute.
    """
    # case sensitive node
    action = InstanceAttributeFactory(Action)


@processor
class ValueRange:

    def __init__(self):
        # attributes are case sensitive node names:
        self.minimum = None
        self.maximum = None
        self.step = None


@processor
class StateVariable:
    """
    Represents a stateVariable with the attributes name, dataType,
    defaultValue, allowedValueList and allowedValueRange.
    """
    # case sensitive node
    allowedValue = ValueSequencer('allowed_values')

    def __init__(self):
        # attributes are case sensitive node names:
        self.name = None
        self.dataType = None
        self.defaultValue = None
        self.allowed_values = list()
        self.allowedValueList = self
        self.allowedValueRange = ValueRange()


@processor
class ServiceStateTable(Storage):
    """
    Collection of stateVariables.
    """
    # case sensitive node
    stateVariable = InstanceAttributeFactory(StateVariable)


class Scpd:
    """
    Provides informations about the Service Control Point Definitions
    for every Service. Every Service has one instance of this class for
    accessing the description of it's own actions and the according
    parameters.
    Root class for processing the content of an scpd-file.
    """
    def __init__(self, root):
        """
        Starts interpreting the scpd-data. 'root' must be an xml.Element
        objects as returned from 'utils.get_xml_root'.
        """
        self._actions = list()
        self._state_variables = list()

        # attributes are case sensitive node names:
        self.specVersion = SpecVersion()
        self.actionList = ActionList(self._actions)
        self.serviceStateTable = ServiceStateTable(self._state_variables)

        # start node processing:
        process_node(self, root)

    @property
    def spec_version(self):
        return self.specVersion.version

    @property
    def actions(self):
        """
        Returns a dictionary with the actions from the actions-list. The
        action-names are the keys and the actions themself are the
        values.
        """
        return {action.name: action for action in self._actions}

    @property
    def state_variables(self):
        """
        Returns a dictionary with the state_variable name as keys and
        the StateVariable itself as value.
        """
        return {sv.name: sv for sv in self._state_variables}


@processor
class Service:
    """
    Class describing a service.
    """
    def __init__(self):
        self._scpd = None
        self._actions = None
        self._state_variables = None
        # attributes are case sensitive node names:
        self.serviceType = None
        self.serviceId = None
        self.controlURL = None
        self.eventSubURL = None
        self.SCPDURL = None

    @property
    def name(self):
        if self.serviceId:
            return self.serviceId.split(':')[-1]
        return None

    @property
    def actions(self):
        """
        Returns all known actions of this service as a dictionary.
        Action names are keys, the action objects are the values. Caches
        the dictionary once retrieved from _scpd.
        """
        if self._actions is None:
            self._actions = self._scpd.actions
        return self._actions

    @property
    def state_variables(self):
        """
        Returns all known stateVariables of this service as a
        dictionary. Names are keys, the stateVariables objects are the
        values. Caches the dictionary once retrieved from _scpd.
        """
        if self._state_variables is None:
            self._state_variables = self._scpd.state_variables
        return self._state_variables

    def load_scpd(self, address, port, timeout=None):
        """Loads the scpd data"""
        protocol = 'http'
        url = f'{protocol}://{address}:{port}{self.SCPDURL}'
        root = get_xml_root(url, timeout=timeout)
        self._scpd = Scpd(root)


@processor
class ServiceList(Storage):
    """
    Collection of Service instances for a device.
    The service instances are stored in the device.services attribute.
    """
    # case sensitive node
    service = InstanceAttributeFactory(Service)


@processor
class Device:
    """
    Storage for devices attributes and device subnodes.
    Subnodes are the serviceList and the deviceList.
    The services provided by a device are collected in services.
    Subdevices are collected in devices.
    All instance attributes are public for read only use.
    """
    def __init__(self):
        self._services = list()
        self.devices = list()
        # attributes are case sensitive node names:
        self.deviceType = None
        self.friendlyName = None
        self.manufacturer = None
        self.manufacturerURL = None
        self.modelDescription = None
        self.modelName = None
        self.modelNumber = None
        self.modelURL = None
        self.UDN = None
        self.UPC = None
        self.presentationURL = None
        self.serviceList = ServiceList(self._services)
        self.deviceList = DeviceList(self.devices)

    @property
    def services(self):
        services = {service.name: service for service in self._services}
        for device in self.devices:
            services.update(device.services)
        return services


@processor
class DeviceList(Storage):
    """
    Collection of sub-devices of a device.
    The Device instances are stored in the device.devices attribute of
    the parent device.
    """
    # case sensitive node
    device = InstanceAttributeFactory(Device)


class Description:
    """
    Root class for a given description information as the content from
    the files igddesc.xml or tr64desc.xml.
    """
    def __init__(self, root):
        """
        Starts data-processing. 'root' must be an xml.Element object as
        returned from 'utils.get_xml_root'.
        """
        # attributes are case sensitive node names:
        self.device = Device()
        self.specVersion = SpecVersion()
        self.systemVersion = SystemVersion()

        # start node processing:
        process_node(self, root)

    @property
    def device_model_name(self):
        return self.device.modelName

    @property
    def spec_version(self):
        return self.specVersion.version

    @property
    def system_version(self):
        """
        Returns the system version of the Fritz!Box as a string like
        '7.10' or None. This information is only available by the
        'tr64desc.xml' file.
        """
        return self.systemVersion.version

    @property
    def system_buildnumber(self):
        """
        Returns the buildnumber or None. This information is only
        available by the 'tr64desc.xml' file.
        """
        return self.systemVersion.Buildnumber

    @property
    def system_display(self):
        """
        Returns the system display-string or None. This information is
        only available by the 'tr64desc.xml' file.
        """
        return self.systemVersion.Display

    @property
    def services(self):
        """
        Returns dictionary with the known services as values and the
        according service-names as keys.
        """
        return self.device.services




import pytest

from ..core.fritzconnection import (
    FritzConnection,
)


@pytest.mark.parametrize(
    "name, expected", [
        ('WANIPConn', 'WANIPConn1'),
        ('WANIPConn1', 'WANIPConn1'),
        ('WANIPConn:1', 'WANIPConn1'),
        ('WANIPConn2', 'WANIPConn2'),
        ('WANIPConn:2', 'WANIPConn2'),
        ('WANIPConn21', 'WANIPConn21'),
        ('WANIPConn:21', 'WANIPConn21'),
    ]
)
def test_normalize_name(name, expected):
    result = FritzConnection.normalize_name(name)
    assert result == expected




import pytest

from ..core.exceptions import (
    FRITZ_ERRORS,
    ActionError,
    ServiceError,
    FritzActionError,
    FritzArgumentError,
    FritzActionFailedError,
    FritzArgumentValueError,
    FritzOutOfMemoryError,
    FritzSecurityError,
    FritzArrayIndexError,
    FritzLookUpError,
    FritzArgumentStringToShortError,
    FritzArgumentStringToLongError,
    FritzArgumentCharacterError,
    FritzInternalError,
)

from ..core.soaper import (
    raise_fritzconnection_error,
    boolean_convert,
)


content_template = """
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <s:Body>
    <s:Fault>
      <faultcode>s:Client</faultcode>
      <faultstring>UPnPError</faultstring>
      <detail>
        <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
          <errorCode>{error_code}</errorCode>
          <errorDescription>Invalid Action</errorDescription>
        </UPnPError>
      </detail>
    </s:Fault>
  </s:Body>
</s:Envelope>
"""


class Response:
    """Namespace object."""


@pytest.mark.parametrize(
    "error_code, exception", [
        ('401', FritzActionError),
        ('402', FritzArgumentError),
        ('501', FritzActionFailedError),
        ('600', FritzArgumentValueError),
        ('603', FritzOutOfMemoryError),
        ('606', FritzSecurityError),
        ('713', FritzArrayIndexError),
        ('714', FritzLookUpError),
        ('801', FritzArgumentStringToShortError),
        ('802', FritzArgumentStringToLongError),
        ('803', FritzArgumentCharacterError),
        ('820', FritzInternalError),

        ('713', IndexError),
        ('714', KeyError),

        ('401', ActionError),

    ]
)
def test_raise_fritzconnection_error(error_code, exception):
    """check for exception raising depending on the error_code"""
    content = content_template.format(error_code=error_code)
    response = Response()
    response.content = content.encode()
    pytest.raises(exception, raise_fritzconnection_error, response)


@pytest.mark.parametrize(
    "value, expected_result", [
        ('0', False),
        ('1', True),
        ('2', True),
        ('x', 'x'),
        ('3.1', '3.1'),
    ]
)
def test_boolean_convert(value, expected_result):
    result = boolean_convert(value)
    assert result == expected_result


long_error = """
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <s:Body>
    <s:Fault>
      <faultcode> s:Client </faultcode>
      <faultstring>
UPnPError </faultstring>
      <detail>
        <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
          <errorCode> 401 </errorCode>
          <errorDescription> Invalid Action </errorDescription>
        </UPnPError>
      </detail>
    </s:Fault>
  </s:Body>
</s:Envelope>
"""


def test_long_error_message():
    response = Response()
    response.content = long_error.encode()
    with pytest.raises(ActionError) as exc:
        raise_fritzconnection_error(response)

    assert exc.value.args[0] == "\n".join(
        ["UPnPError: ",
         "errorCode: 401",
         "errorDescription: Invalid Action",
         ]
    )



"""
Common functions for other core-modules.
"""

import re
import requests

from xml.etree import ElementTree as etree
from .exceptions import FritzConnectionException


NS_REGEX = re.compile("({(?P<namespace>.*)})?(?P<localname>.*)")


def localname(node):
    if callable(node.tag):
        return "comment"
    m = NS_REGEX.match(node.tag)
    return m.group('localname')


def get_content_from(url, timeout=None):
    conn = requests.get(url, timeout=timeout)
    ct = conn.headers.get("Content-type")
    if ct == "text/html":
        raise FritzConnectionException("Unable to login into device to get configuration information.")
    return conn.text


def get_xml_root(source, timeout=None):
    """
    Function to help migrate from lxml to the standard-library xml-package.

    'source' must be a string and can be an xml-string, a uri or a file
    name. `timeout` is an optional parameter limiting the time waiting
    for a router response.
    In all cases this function returns an xml.etree.Element instance
    which is the root of the parsed tree.
    """
    if source.startswith("http://") or source.startswith("https://"):
        # it's a uri, use requests to get the content
        source = get_content_from(source, timeout=timeout)
    elif not source.startswith("<"):
        # assume it's a filename
        with open(source) as fobj:
            source = fobj.read()
    return etree.fromstring(source)




import argparse

from .. import package_version
from ..lib.fritzhomeauto import FritzHomeAutomation
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)


def report_verbose(fh):
    informations = fh.device_informations()
    for info in informations:
        width = len(max(info.keys(), key=lambda x: len(x)))
        line = f'{{attribute:{width}}} : {{value}}'
        for attribute in sorted(info.keys()):
            print(line.format(attribute=attribute, value=info[attribute]))
        print()  # add blank line between devices


def report_compact(fh):
    name = 'Device Name'
    ain = 'AIN'
    power = 'Power[W]'
    temperature = 't[°C]'
    switch_state = 'switch'
    print(f'{name:24}{ain:18}{power:>10}{temperature:>8}   {switch_state}')
    for di in fh.device_informations():
        name = di['NewDeviceName']
        ain = di['NewAIN']
        ain = f"'{ain}'"
        power = di['NewMultimeterPower'] * 0.01
        temperature = di['NewTemperatureCelsius'] *0.1
        switch_state = di['NewSwitchState'].lower()
        print(f'{name:24}{ain:18}{power:>10.3f}{temperature:>8.1f}   {switch_state}')


def report_status(fh, arguments):
    print(f'\nFritzConnection v{package_version}')
    print(fh.fc)
    print('Status of registered home-automation devices:\n')
    if arguments.verbose:
        report_verbose(fh)
    else:
        report_compact(fh)


def switch_device(fh, arguments):
    ain = arguments.switch[0]
    state = arguments.switch[1].lower() == 'on'
    fh.set_switch(identifier=ain, on=state)


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='FritzBox HomeAuto')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help='ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help='port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    parser.add_argument('-v', '--verbose',
                        nargs='?', default=False, const=True,
                        help='report in verbose mode')
    parser.add_argument('-s', '--switch',
                        nargs=2,
                        help='set switch state. requires two parameters: '
                             'ain and state [on|off]')
    args = parser.parse_args()
    return args


def main():
    arguments = get_cli_arguments()
    if not arguments.password:
        print('Exit: password required.')
        return
    fh = FritzHomeAutomation(address=arguments.address,
                             port=arguments.port,
                             user=arguments.username,
                             password=arguments.password)
    if arguments.switch:
        switch_device(fh, arguments)
    else:
        report_status(fh, arguments)


if __name__ == '__main__':
    main()



from io import StringIO
from pathlib import Path
import os

import pytest
from xml.etree import ElementTree as etree

from ..core.devices import DeviceManager

@pytest.mark.parametrize("filename", [
    "igddesc.xml",
    "tr64desc.xml",
]
                         )
def test_DeviceManager_load_descriptions(device_manager, datadir, filename):
    """test loading of two description files."""
    os.chdir(datadir)
    device_manager.add_description(filename)
    assert device_manager.modelname == 'FRITZ!Box 7590'


@pytest.fixture()
def device_manager(datadir):
    os.chdir(datadir)
    manager = DeviceManager()
    manager.add_description('igddesc.xml')
    manager.add_description('tr64desc.xml')
    return manager


@pytest.mark.parametrize(
    "name", ['WANIPConn1', 'DeviceConfig1', 'WLANConfiguration1']
)
def test_DeviceManager_get_service(name, device_manager):
    """
    test services dict of the DeviceManager. Should find nested services
    (from sub devices) and from different files.
    """
    device_manager.scan()
    # should not raise a KeyError
    service = device_manager.services[name]
    assert service.name == name



"""
fritzhosts.py

Module to inspect the FritzBox API for registered hosts.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""


import argparse

from .. import package_version
from ..lib.fritzhosts import FritzHosts
from ..core.fritzconnection import (
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)


def print_status(arguments):
    print(f'\nFritzConnection v{package_version}')
    fh = FritzHosts(address=arguments.address,
                    port=arguments.port,
                    user=arguments.username,
                    password=arguments.password)
    print(f'FritzHosts for {fh.fc}:\n')
    print('List of registered hosts:\n')
    print('{:>3}: {:<16} {:<28} {:<17}   {}\n'.format(
        'n', 'ip', 'name', 'mac', 'status'))
    hosts = fh.get_hosts_info()
    for index, host in enumerate(hosts, start=1):
        status = 'active' if host['status'] else  '-'
        ip = host['ip'] if host['ip'] else '-'
        mac = host['mac'] if host['mac'] else '-'
        hn = host['name']
        print(f'{index:>3}: {ip:<16} {hn:<28} {mac:<17}   {status}')
    print('\n')


def get_cli_arguments():
    parser = argparse.ArgumentParser(description='FritzBox Hosts')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=None, const=None,
                        dest='address',
                        help='ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        dest='port',
                        help='port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=None, const=None,
                        help='Fritzbox authentication password')
    args = parser.parse_args()
    return args


def main():
    arguments = get_cli_arguments()
    if not arguments.password:
        print('Exit: password required.')
    else:
        print_status(arguments)


if __name__ == '__main__':
    main()



"""
fritzinspection.py

Module to inspect the FritzBox API for available services and actions.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""


import argparse
import os

from ..core.exceptions import FritzConnectionException
from ..core.fritzconnection import (
    FritzConnection,
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
)
from .. import package_version


class FritzInspection:
    """
    Class for cli use to inspect available services and according
    actions of the given device (the Fritz!Box).
    """
    # pylint: disable=invalid-name  # self.fc is ok.

    def __init__(self, address, port, user, password):
        self.fc = FritzConnection(address, port, user, password)

    def view_header(self):
        print(self.fc)

    def view_servicenames(self):
        """Send all known service names to stdout."""
        print('Servicenames:')
        for service_name in self.fc.services:
            print('{:20}{}'.format('', service_name))

    def view_actionnames(self, service_name, view_arguments=False):
        """Send all action names of the given service to stdout."""
        print('\n{:<20}{}'.format('Servicename:', service_name))
        print('Actionnames:')
        try:
            service = self.fc.services[service_name]
        except KeyError as err:
            print(f'Error: Invalid Servicename {err}')
        else:
            for action_name in service.actions:
                print('{:20}{}'.format('', action_name))
                if view_arguments:
                    action = service.actions[action_name]
                    for argument in sorted(action.arguments.keys()):
                        print('{:24}- {}'.format('', argument))
                    print()


    def view_actionarguments(self, service_name, action_name):
        """Send all action names of the given service to stdout."""
        try:
            service = self.fc.services[service_name]
        except KeyError as err:
            print(f'Error: Invalid Servicename {err}')
            return
        try:
            action = service.actions[action_name]
        except KeyError as err:
            print(f'Error: Invalid Actionname {err}')
            return
        print('\n{:<20}{}'.format('Service:', service_name))
        print('{:<20}{}'.format('Action:', action_name))
        print('Parameters:\n')
        print('    {:30}{:14}{}\n'.format('Name', 'direction', 'data type'))
        for argument in action.arguments.values():
            if argument.direction == 'in':
                direction = '-> in'
            else:
                direction = '   out ->'
            var = service.state_variables.get(argument.relatedStateVariable, '')
            line = f'    {argument.name:30}{direction:14}{var.dataType}'
            print(line)


def get_cli_arguments():
    """
    Returns a NameSpace object from the ArgumentParser parsing the given
    command line arguments.
    """
    print(f'\nFritzConnection v{package_version}')
    parser = argparse.ArgumentParser(description='Fritz!Box API Inspection:')
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=FRITZ_IP_ADDRESS, const=None,
                        dest='address',
                        help='Specify ip-address of the FritzBox to connect to.'
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        help='Port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=os.getenv('FRITZ_USERNAME', None),
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=os.getenv('FRITZ_PASSWORD', None),
                        help='Fritzbox authentication password')
    parser.add_argument('-r', '--reconnect',
                        action='store_true',
                        help='Reconnect and get a new ip')
    parser.add_argument('-s', '--services',
                        action='store_true',
                        help='List all available services')
    parser.add_argument('-S', '--serviceactions',
                        nargs=1,
                        help='List actions for the given service: <service>')
    parser.add_argument('-a', '--servicearguments',
                        nargs=1,
                        help='List arguments for the actions of a '
                             'specified service: <service>.')
    parser.add_argument('-A', '--actionarguments',
                        nargs=2,
                        help='List arguments for the given action of a '
                             'specified service: <service> <action>. '
                             'Lists also direction and data type of the '
                             'arguments.')
    args = parser.parse_args()
    return args


def get_inspector(args):
    try:
        inspector = FritzInspection(
            args.address, args.port, args.username, args.password)
    except FritzConnectionException as err:
        return None
    return inspector


def run_inspector(inspector, args):
    inspector.view_header()
    if args.services:
        inspector.view_servicenames()
    elif args.serviceactions:
        inspector.view_actionnames(args.serviceactions[0])
    elif args.servicearguments:
        inspector.view_actionnames(args.servicearguments[0], view_arguments=True)
    elif args.actionarguments:
        inspector.view_actionarguments(args.actionarguments[0],
                                       args.actionarguments[1])
    elif args.reconnect:
        inspector.fc.reconnect()
        print('reconnect the router.')


def main():
    """CLI entry point."""
    args = get_cli_arguments()
    inspector = get_inspector(args)
    if inspector:
        run_inspector(inspector, args)
    print() # print an empty line


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
from fritzconnection import package_version


with open('README.rst') as file:
    the_long_description = file.read()


setup(
    name = 'fritzconnection',
    version = package_version,
    packages = find_packages(),
    license = 'MIT',
    description = 'Communicate with the AVM FritzBox',
    long_description = the_long_description,
    author = 'Klaus Bremer',
    author_email = 'bremer@bremer-media.de',
    url = 'https://github.com/kbr/fritzconnection',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords = 'AVM FritzBox',
    python_requires = ">= 3.6",
    install_requires = [
        'requests>=2.22.0',
    ],
    entry_points={'console_scripts': [
        'fritzconnection = fritzconnection.cli.fritzinspection:main',
        'fritzcall = fritzconnection.cli.fritzcall:main',
        'fritzhomeauto = fritzconnection.cli.fritzhomeauto:main',
        'fritzhosts = fritzconnection.cli.fritzhosts:main',
        'fritzphonebook = fritzconnection.cli.fritzphonebook:main',
        'fritzstatus = fritzconnection.cli.fritzstatus:main',
        'fritzwlan = fritzconnection.cli.fritzwlan:main',
    ]}
)



"""
Module handling the SOAP based communication with the router.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import datetime
import re

import requests
from requests.auth import HTTPDigestAuth
from xml.etree import ElementTree as etree

from .exceptions import (
    FritzConnectionException,
    FRITZ_ERRORS,
)
from .utils import localname


SOAP_NS = "http://schemas.xmlsoap.org/soap/envelope/"


def datetime_convert(value):
    """Converts a string in ISO 8601 format to a datetime-object."""
    try:
        return datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
    except ValueError:
        return value


def boolean_convert(value):
    """Converts a string like '1' or '0' to a boolean value"""
    try:
        return bool(int(value))
    except ValueError:
        # should not happen: leave value as is
        return value


def uuid_convert(value):
    """Strips the leading 'uuid:' part from the string."""
    return value.split(':')[-1]


def raise_fritzconnection_error(response):
    """
    Handles all responses with status codes other than 200.
    Will raise the relevant FritzConnectionException with
    the error code and description if available
    """
    parts = []
    error_code = None
    try:
        root = etree.fromstring(response.content)
    except etree.ParseError:
        # May fail in case it's html instead of xml.
        # Can happen on wrong authentication.
        # That means it is not an error reported from executing
        # some service in the box, but rather not allowed to
        # access the box at all.
        # Whatever it is, report it here:
        detail = re.sub(r'<.*?>', '', response.text)
        msg = f'Unable to perform operation. {detail}'
        raise FritzConnectionException(msg)
    detail = root.find('.//detail')
    children = detail.iter()
    next(children) # skip detail itself
    for node in children:
        tag = localname(node)
        text = node.text.strip()
        if tag == "errorCode":
            error_code = text
        parts.append(f'{tag}: {text}')
    message = '\n'.join(parts)
    # try except:KeyError not possible,
    # because one of the raised Exceptions may inherit from KeyError.
    exception = FRITZ_ERRORS.get(error_code, FritzConnectionException)
    raise exception(message)


class Soaper:
    """
    Class making the soap on its own to communicate with the FritzBox.
    Instead of ham, spam and eggs, it's hopelessly addicted to soap.

    For accessing the Fritz!Box the parameters `address` for the router
    ip, `port`, `user` and `password` are required. The optional
    parameter `timeout` will limit the time Soaper waits for a router
    response. (These parameters will get set by FritzConnection,)
    """

    headers = {
        'soapaction': '',
        'content-type': 'text/xml',
        'charset': 'utf-8'
    }

    envelope = re.sub(r'\s +', '', """
        <?xml version="1.0" encoding="utf-8"?>
        <s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"
                    xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">{body}
        </s:Envelope>
        """)

    body_template = re.sub(r'\s +', '', """
        <s:Body>
        <u:{action_name} xmlns:u="{service_type}">{arguments}
        </u:{action_name}>
        </s:Body>
        """)

    argument_template = "<s:{name}>{value}</s:{name}>"
    method = 'post'

    conversion_table = {
        'datetime': datetime_convert,
        'boolean': boolean_convert,
        'uuid': uuid_convert,
        'i4': int,
        'ui1': int,
        'ui2': int,
        'ui4': int,
    }

    def __init__(self, address, port, user, password, timeout=None):
        self.address = address
        self.port = port
        self.user = user
        self.password = password
        self.timeout = timeout

    def get_body(self, service, action_name, arguments):
        """Returns the body by template substitution."""
        return self.body_template.format(
            service_type=service.serviceType,
            action_name=action_name,
            arguments=arguments
        )

    def execute(self, service, action_name, arguments):
        """
        Builds the soap request and returns the response as dictionary.
        Numeric and boolean values are converted from strings to Python
        datatypes.
        """
        headers = self.headers.copy()
        headers['soapaction'] = f'{service.serviceType}#{action_name}'
        arguments = ''.join(self.argument_template.format(name=k, value=v)
                            for k, v in arguments.items())
        body = self.get_body(service, action_name, arguments)
        envelope = self.envelope.format(body=body)
        protocol = 'http'
        url = f'{protocol}://{self.address}:{self.port}{service.controlURL}'
        auth = None
        if self.password:
            auth = HTTPDigestAuth(self.user, self.password)
        response = requests.post(url, data=envelope, headers=headers,
                                      auth=auth, timeout=self.timeout)
        if response.status_code != 200:
            raise_fritzconnection_error(response)
        return self.parse_response(response, service, action_name)

    def parse_response(self, response, service, action_name):
        """
        Extracts all known parameters of the given action from the
        response and returns this as a dictionary with the out-parameter
        names as keys and the corresponding response as values.
        Will raise an ActionError on unknown action_name.
        """
        result = dict()
        action = service.actions[action_name]
        root = etree.fromstring(response.content)
        for argument_name in action.arguments:
            try:
                value = root.find(f'.//{argument_name}').text
            except AttributeError:
                continue
            state_variable_name = \
                action.arguments[argument_name].relatedStateVariable
            state_variable = service.state_variables[state_variable_name]
            data_type = state_variable.dataType.lower()
            try:
                value = self.conversion_table[data_type](value)
            except KeyError:
                pass
            result[argument_name] = value
        return result
